#include "generic.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <math.h>

#include "vektorr3.h"

using namespace std;

VektorR3 span_x, span_y;
VektorR3 center;

VektorR3 observer;
VektorR3 target;

VektorR3 sight;
VektorR3 light;

double   refract_i;
double   refract_t;

SDL_Surface *screen;
int          window_width, window_height;
double       obs_angle = 8.25;

inline double sqr(double v)
{
  return v*v;
} // srq

inline void SetPixelColor(SDL_Surface *screen, int x, int y, unsigned char red, unsigned char green, unsigned char blue)
{
  ((unsigned char*)(screen->pixels))[x*3+0 + y*screen->pitch] = blue;
  ((unsigned char*)(screen->pixels))[x*3+1 + y*screen->pitch] = green;
  ((unsigned char*)(screen->pixels))[x*3+2 + y*screen->pitch] = red;
} // SetPixelColor

inline void SetPixel(SDL_Surface* screen, int x, int y)
{
  SetPixelColor(screen, x, y, 255, 255, 255);
} // SetPixel

bool CalcProPlane(VektorR3 B, VektorR3 Z, int Res_x, int Res_y, int angle,  VektorR3 &A, VektorR3 &x, VektorR3 &y)
{
	VektorR3 b;
	float    laenge;

	// Blickvektor berechnen

	b = Z - B;

	if((b[0] != 0.0) || (b[1] != 0.0) || (b[2] != 0.0))
	{
		// Aufhnger A der Ebene berechnen

		laenge = b.Norm2();

		A = B + b/laenge;

		// Berechnung des Spannvektors x

		x[1] = 0.0; // Kein geneigter Betrachter -> Projektionsebene parallel zur X-Z Ebene

		if(b[2] != 0.0)
		{
			x[0] = 1/((float)sqrt((double)(1.0+sqr(b[0]/b[2]))));
			x[2] = (-x[0]*b[0])/b[2];
		} // if
		else
		{
			if(b[0] != 0.0)
			{
				x[0] = 0.0;
				x[2] = 1.0;
			} // if
			else
			{
				x[0] = 1.0;
				x[2] = 0.0;
			} // else
		} // else

		// Berechnung des Spannvektors y

		if(x[2] != 0.0)
		{
			if(b[1] != 0.0)
			{
				y[0] = 1.0/((float)sqrt((double)(1.0+sqr((x[0]*b[2])/(x[2]*b[1])-b[0]/b[1])+sqr(x[0]/x[2]))));
				y[1] = y[0]*((x[0]*b[2])/(x[2]*b[1])-b[0]/b[1]);
				y[2] = (-x[0]*y[0])/x[2];
			} // if
			else
			{
			  y = VektorR3(0.0, 1.0, (-x[0]*y[0])/x[2]);
			} // else
		} // if
		else
		{
			if(b[2] != 0.0)
			{
				y[1] = 1.0/((float)sqrt((double)(1.0+sqr(b[1]/b[2]))));
				y[2] = (-y[1]*b[1])/b[2];
				y[0] = 0.0;
			} // if
			else
			{
				y = VektorR3(0,1,0);
			} // else
		} // else

		// Vektorrichtung bestimmen

		if(y[1] < 0.0)
		{
			y *= -1.0;
		} // if

		if(b[2] <= 0.0)
		{
			if(x[0] < 0.0)
			{
				x *= -1.0;
			} // if
		} // if
		else
		{
			if(x[0] > 0.0)
			{
				x *= -1.0;
			} // if
		} // else

		// Berechnung der Vektorlnge

		laenge = ((float)tan((double)((float)angle/114.591559))*2.0)/(float)Res_x;

		x *= laenge;
		y *= laenge;

		return(true);
	} // if
	else
	  return(false); // Betrachter = Ziel (darf nicht sein)
} // CalcProPlane

VektorR3 CalcReflexion(const VektorR3& in, const VektorR3& normal)
{
  VektorR3 reflexion;

  double length = normal.Norm2();

  double skalar = -2*in*normal/length;

  reflexion = in + skalar*normal/length;

  return reflexion;
} // CalcReflexion

// return the parameter of the line through center of projection and the
// offset (midpoint of projection plane). 1 = intersection right at projection
// plane, -1 indicates no intersection, values > 1 indicate intersections
// beyond the intersection plane

double IntersectSphere(const VektorR3& center, double radius, VektorR3& offset, VektorR3& direction)
{
  double s, s1, s2;
  double v;
  double a, b, c;

  VektorR3 S = offset - center;

  a = direction*direction;
  b = 2*direction*S;
  c = S*S-sqr(radius);

  v = sqr(b) - 4*a*c;

  if(v < 0.0) return -1;

  s1 = (-b + sqrt(v))/(2*a);
  s2 = (-b - sqrt(v))/(2*a);

  s = s1;
  if((s2 < s) && (s2 > 0)) s = s2;

  // if(s < 0.01) s = -1; // prevent self-cut

  return s;
} // IntersectSphere

double IntersectTriangle(const VektorR3 P[3], VektorR3& offset, VektorR3& direction, VektorR3& cut, VektorR3& n, double& span_a, double& span_b)
{
  double s;
  int    i0, i1;

  VektorR3 a = P[1] - P[0];
  VektorR3 b = P[2] - P[0];

  n = VektorR3::CrossProduct(a, b);
  VektorR3 v = offset - P[0];

  s = -(n*v)/(n*direction);

  if(s > 0.0) {

    cut = offset + s*direction;

    for(int index = 0; index < 3; index++) {
    
      switch(index) {
      case 0:
	i0=0; i1=1;
	break;
	
      case 1:
	i0=0; i1=2;
	break;
	
      case 2:
	i0=1; i1=2;
	break;
	
      } // switch
      
      double denom = (a[i0]*b[i1]+a[i1]*b[i0]);
      
      if(fabs(denom) > 0.01) {
	
	span_b = (a[i0]*(cut[i1] - P[0][i1]) - a[i1]*(cut[i0] - P[0][i0])) / denom;
	
	if((span_b >= 0.0) && (span_b <= 1.0)) {
	  span_a = (cut[i0] - P[0][i0] - span_b*b[i0])/a[i0];
	  if((span_a >= 0.0) && (span_a <= 1.0))
	    if(span_a+span_b <= 2.0) return s;
	  
	} // if
      } // if
      
    } // for
  } // if
  
  return -1;
    
} // IntersectTriangle

void DetermineColor(VektorR3 &observer, VektorR3 &sight, unsigned char &red, unsigned char &green, unsigned char &blue, int recursion = 1)
{
  VektorR3 reflexion, intersection, incident_light, surf_normal, gaze;
  double   s1, s2, specular, diffuse, illumination, shade;
  double   diffuse_magnitude;
  double   span_a, span_b;
  double   highlight_focus, highlight_strength;

  int      obj_color_red, obj_color_green, obj_color_blue;
  int      add_color_red, add_color_green, add_color_blue;

  VektorR3 n, cut;

  VektorR3 light(60, -100, -50);

  VektorR3 P[2][3];
  VektorR3 mid_sphere;

  red = green = blue = 0;
  add_color_red = add_color_green = add_color_blue = 0;

  if(recursion > 4) return;

  P[0][0] = VektorR3(-100, 60, 100);
  P[0][1] = VektorR3(80, 60, 100);
  P[0][2] = VektorR3(-100, 60, 250);
  
  P[1][0] = VektorR3(-100, 60, 250);
  P[1][1] = VektorR3(80, 60, 250);
  P[1][2] = VektorR3(-100, -60, 250);
  
  mid_sphere = VektorR3(15, 20, 200);

  s1 = IntersectSphere(mid_sphere, 45.0, observer, sight);
  s2 = IntersectTriangle(P[0], observer, sight, cut, n, span_a, span_b);
  
  if(s2 == -1)
    s2 = IntersectTriangle(P[1], observer, sight, cut, n, span_a, span_b);
  
  if((s1 > 0) && ((s1 < s2) || (s2 < 0))) {

    // hier wird die kugel berechnet

    cut = observer + s1*sight;

    n = cut - mid_sphere;

    obj_color_red = (int)255;
    obj_color_green = (int)180;
    obj_color_blue = (int)5;

    diffuse_magnitude = 0.8;

    add_color_red = 0;
    add_color_green = 0;
    add_color_blue = 0;
   
    // calculcate reflexion

    VektorR3      obs, sig;
    unsigned char sec_color_red, sec_color_green, sec_color_blue;

    sig = CalcReflexion(sight, n);

    VektorR3 cut2 = cut + 0.001*sig;

    DetermineColor(cut2, sig, sec_color_red, sec_color_green, sec_color_blue, recursion+1);
    
    add_color_red += (int)sec_color_red*0.35;
    add_color_green += (int)sec_color_green*0.35;
    add_color_blue += (int)sec_color_blue*0.35;

    cut2 = cut + 0.001*sight;

    DetermineColor(cut2, sight, sec_color_red, sec_color_green, sec_color_blue, recursion+1);
    
    add_color_red += (int)sec_color_red*0.35;
    add_color_green += (int)sec_color_green*0.35;
    add_color_blue += (int)sec_color_blue*0.35;

    highlight_focus = 128; highlight_strength = 50;

  } // if

  if((s2 > 0) && ((s2 < s1) || (s1 < 0))) {

    cut = observer + s2*sight;

    diffuse_magnitude = 1.0;

    int myval = (int)((int)(span_b*40)/5)*5; 

    //    cout << (int)(span_b*10) << endl;

    if((int)((span_a)*50+myval) % 10 < 5) {
      obj_color_red = 64;
      obj_color_green = 64;
      obj_color_blue = 64;
    } // if
    else {
      obj_color_red = 128;
      obj_color_green = 128;
      obj_color_blue = 135;
    } // else

    highlight_focus = 128; highlight_strength = 1;
  } // if

  if((s1 < 0) && (s2 < 0)) {
    sight *= 1.0/sight.Norm2();
    shade = (0.0+pow((sight*VektorR3(0,-1.0,0)), 4))*5;
    if(shade > 1.0) shade = 1.0;
    
    red = 0;
    green = 0;
    double fblue = (double)(shade*255.0+48);
    
    if(fblue > 255) fblue = 255;
    
    blue = fblue;
    
    return;
  } // if

  n *= 1.0/n.Norm2();
  
  VektorR3 l = light - cut;
  l *= 1.0/l.Norm2();
  
  // determine diffuse color
  
  shade = l*n;
  
  // determine specular reflexion
  
  VektorR3 r = (2.0*n*l)*n - l;
  r *= 1.0/r.Norm2();
  
  sight *= 1.0/sight.Norm2();
  
  double spec = -1.0*r*sight;
  
  if(spec > 0.0) shade += pow(spec,highlight_focus)*highlight_strength;
  
  if(shade < 0.0) shade = 0.0;

  shade += 0.09; // some ambient light exists everywhere
  
  //  if(shade > 1.0) shade = 1.0;

  int fr, fg, fb;

  fr = shade*obj_color_red*diffuse_magnitude + add_color_red;
  fg = shade*obj_color_green*diffuse_magnitude + add_color_green;
  fb = shade*obj_color_blue*diffuse_magnitude + add_color_blue;

  if(fr > 255) fr = 255;
  if(fg > 255) fg = 255;
  if(fb > 255) fb = 255;
  
  red = fr;
  green = fg;
  blue = fb;

  return;
} // DetermineColor    

void RenderImage(VektorR3 observer, VektorR3 target, int delta_x)
{
  unsigned char red, green, blue;

  CalcProPlane(observer, target, window_width, window_height, 
	       60, center, span_x, span_y);
    
  for(int y = 0; y < window_height; y++) {
    for(int x = 0; x < window_width; x++) {
      
      sight = (center + ((double)(x-window_width/2)*span_x + (double)(y-window_height/2)*span_y)) - observer;
      
      DetermineColor(observer, sight, red, green, blue);
      
      SetPixelColor(screen, x+delta_x, y, red, green, blue);
      
    } // for		    
  } // for  
} // RenderImage

int main(int argc, char** argv)
{
    SDL_Event    event;
    SDL_Rect     dst_rect, full_image_rect, scrolled_rect, location;
    long         pixel_set = 0;

    refract_i = 3.14/2;
    refract_t = 3.14/4;

    window_width = 350; window_height = 350;

    if(SDL_Init(SDL_INIT_VIDEO) < 0) {
	cerr << "Could not initialize SDL video" << endl;
	exit(-1);
    } // if
    atexit(SDL_Quit);

    Uint32 flags = 0;

    // flags |= SDL_FULLSCREEN;

    screen = SDL_SetVideoMode(window_width*2, window_height, 24, flags);

    if(!screen) {
	cerr << "could not SetVideoMode() - abort" << endl;
	exit(-1);
    } // if

    for(;;) {
      
      while(SDL_WaitEvent(&event) == 0) {
      } // while
      /*
	if(( == 0) && (event.type == SDL_KEYDOWN)) 
	    goto exit_app;
      */
	switch(event.type) {
		
	    case SDL_MOUSEMOTION:	      
		break;
		
	    case SDL_MOUSEBUTTONDOWN:
		break;
		
	    case SDL_MOUSEBUTTONUP:
		break;
		
	    case SDL_KEYDOWN:
		
		switch(event.key.keysym.sym) {
		
		case '+':
		  refract_t += 0.01;
		  break;
		  
		case '-':
		  refract_i += 0.01;
		  break;
		    
		case ' ':
		  observer = VektorR3(sin(obs_angle)*90, 0, cos(obs_angle)*90);
		  target = VektorR3(0, 0, 190);

		  RenderImage(observer, target, 0);

		  observer = VektorR3(sin(obs_angle+0.2)*90, 0, cos(obs_angle)*90);
		  target = VektorR3(0, 0, 190);

		  RenderImage(observer, target, window_width);

		  obs_angle += 0.25;

		  SDL_UnlockSurface(screen);
		  SDL_UpdateRect(screen, 0, 0, window_width*2, window_height);
		  SDL_LockSurface(screen);
		  break;

		case 'q':
		case 'Q':
		  goto exit_app;
			
		default:
		  break;
		} // switch
	} // switch

    } // for
	    
 exit_app:

    SDL_UnlockSurface(screen);	    
    SDL_FreeSurface(screen);

    exit(0);
} // main
